#include "ofApp.h"
#include "aeroTAP/aeroTAPCamera.h"
#include "aeroTAP/aeroTAPLinuxCamera.h"
#include "aeroTAP/ZDColorPalette.h"


// #define _GS is defined in aeroTAPCamera.h

using namespace aeroTAP;

typedef unsigned short WORD;

aeroTAPLinux3DCamera *aeroCam;
ZDColotPalette ColorPalette;

int nZDTable[2048];
int nZDTableLen = 0;

ofImage depthMap;
ofImage colorImg;

void YUY2_to_RGB24(unsigned char *pSrc, unsigned char *pDst, int cx, int cy)
{
        int nSrcBPS,nDstBPS,x,y,x2,x3,m;
        int ma0,mb0,m02,m11,m12,m21;
        unsigned char *pS0,*pD0;
        int Y,U,V,Y2;
        unsigned char R,G,B,R2,G2,B2;
        //
        nSrcBPS = cx*2;
        nDstBPS = ((cx*3+3)/4)*4;
        //
        pS0 = pSrc;
#if 1
        pD0 = pDst;
#else
        pD0 = pDst+nDstBPS*(cy-1);
#endif
        for (y=0; y<cy; y++) {
                for (x3=0,x2=0,x=0; x<cx; x+=2,x2+=4,x3+=6) {
                        Y = (int)pS0[x2+0]-16;
                        Y2= (int)pS0[x2+2]-16;
                        U = (int)pS0[x2+1]-128;
                        V = (int)pS0[x2+3]-128;
                        //
                        ma0 = 298*Y;
                        mb0 = 298*Y2;
                        m02 =  409*V+128;
                        m11 = -100*U;
                        m12 = -208*V+128;
                        m21 =  516*U+128;
                        //
                        m = (ma0    +m02)>>8;
                        R = (m<0)?(0):(m>255)?(255):((unsigned char)m);
                        m = (ma0+m11+m12)>>8;
                        G = (m<0)?(0):(m>255)?(255):((unsigned char)m);
                        m = (ma0+m21    )>>8;
                        B = (m<0)?(0):(m>255)?(255):((unsigned char)m);
                        //
                        m = (mb0    +m02)>>8;
                        R2 = (m<0)?(0):(m>255)?(255):((unsigned char)m);
                        m = (mb0+m11+m12)>>8;
                        G2 = (m<0)?(0):(m>255)?(255):((unsigned char)m);
                        m = (mb0    +m21)>>8;
                        B2 = (m<0)?(0):(m>255)?(255):((unsigned char)m);
                        //
 #if 1
                        pD0[x3+2] = B;
                        pD0[x3+1] = G;
                        pD0[x3+0] = R;
                        pD0[x3+5] = B2;
                        pD0[x3+4] = G2;
                        pD0[x3+3] = R2;
#else
                        pD0[x3  ] = B;
                        pD0[x3+1] = G;
                        pD0[x3+2] = R;
                        pD0[x3+3] = B2;
                        pD0[x3+4] = G2;
                        pD0[x3+5] = R2;
#endif
                }
                pS0 += nSrcBPS;
#if 1
                pD0 += nDstBPS;
#else
                pD0 -= nDstBPS;
#endif
        }
}
void loadZDTable(string filename) {
	ifstream f;
	f.open(ofToDataPath(filename).c_str());
	if (!f.fail())
	{
		nZDTableLen = 0;
		vector <string> strLines;
		while (!f.eof()) {
			string ptStr;
			getline(f, ptStr);
			if (ptStr.length()<2)
				continue;
			int val = std::atoi(ptStr.c_str());
			nZDTable[nZDTableLen] = val;
			++nZDTableLen;
		}
		f.close();
	}
	printf("Loaded ZDTable %s %d\n", filename.c_str(), nZDTableLen);
}

//--------------------------------------------------------------
void ofApp::setup(){

	ofSetFrameRate(30);
    camWidth = 640;  // try to grab at this size.
    camHeight = 480;
#if 1
#ifdef _GS
    aeroCam = new aeroTAPLinux3DCamera(camWidth,camHeight,camWidth,camHeight);
    depthMap.allocate(camWidth,camHeight,OF_IMAGE_COLOR);
#else
    aeroCam = new aeroTAPLinux3DCamera(camWidth,camHeight,camWidth/2,camHeight);
    depthMap.allocate(camWidth,camHeight,OF_IMAGE_COLOR);
#endif
    colorImg.allocate(camWidth,camHeight,OF_IMAGE_COLOR);
#else
    //get back a list of devices.
    vector<ofVideoDevice> devices = vidGrabber.listDevices();

    for(size_t i = 0; i < devices.size(); i++){
        if(devices[i].bAvailable){
            //log the device
            ofLogNotice() << devices[i].id << ": " << devices[i].deviceName;
        }else{
            //log the device and note it as unavailable
            ofLogNotice() << devices[i].id << ": " << devices[i].deviceName << " - unavailable ";
        }
    }
    vidGrabber.setDeviceID(0);
    vidGrabber.setDesiredFrameRate(30);
    vidGrabber.initGrabber(camWidth, camHeight);
    videoInverted.allocate(camWidth, camHeight, OF_PIXELS_RGB);
    videoTexture.allocate(videoInverted);
#endif
#ifdef _GS
	loadZDTable(ofToDataPath("zdTable_HD.txt"));
#else
	loadZDTable(ofToDataPath("zdTable_VGA.txt"));
#endif

    ofSetVerticalSync(true);
}


//--------------------------------------------------------------
void ofApp::update(){

//printf("update\n");
//    ofBackground(100, 100, 100);
#if 1
	RGBQUAD *pColorPalette = ColorPalette.GetColorPalette();

    aeroCam->updateFrame();
    emCamStreamFormat typeC =emCamStreamFormatColor;
    unsigned char *pColor = aeroCam->getFrame(typeC);

    emCamStreamFormat typeD =emCamStreamFormatDepth;
    unsigned char *pDepth = aeroCam->getFrame(typeD);

    if ( pColor )
    {
//printf("new frame\n");
        unsigned char *pTrg= colorImg.getPixels().getData();

        YUY2_to_RGB24(pColor, pTrg, camWidth, camHeight);
		colorImg.update();
	}

    if ( pDepth )
    {
#ifdef _GS
	WORD *pDepthW = (WORD*)pDepth;
	unsigned char *pTrg= depthMap.getPixels().getData();
	for(int i=0;i<camWidth*camHeight;++i)
	{
		int nDepth = pDepthW[i];
		if (nZDTableLen > 0)
		{
			if (nDepth < nZDTableLen)
				nDepth = nZDTable[nDepth & 0x3FF];
			else
				nDepth = 0;
			pTrg[i * 3] = pColorPalette[nDepth].rgbRed;
			pTrg[i * 3 + 1] = pColorPalette[nDepth].rgbGreen;
			pTrg[i * 3 + 2] = pColorPalette[nDepth].rgbBlue;
		}
		else
		{
			pTrg[i * 3] = nDepth & 0xFF;
			pTrg[i * 3 + 1] = (nDepth >>8 ) & 0xFF;
			pTrg[i * 3 + 2] = 100;
		}
	}
#else
	unsigned char *pTrg= depthMap.getPixels().getData();
	for(int i=0;i<camWidth*camHeight;++i)
	{
		int nDepth = pDepth[i];
		if (nZDTableLen > 0)
		{
			if (nDepth < nZDTableLen)
				nDepth = nZDTable[nDepth];
			else
				nDepth = 0;
			pTrg[i * 3] = pColorPalette[nDepth].rgbRed;
			pTrg[i * 3 + 1] = pColorPalette[nDepth].rgbGreen;
			pTrg[i * 3 + 2] = pColorPalette[nDepth].rgbBlue;
		}
		else
		{
			pTrg[i * 3] = nDepth;
			pTrg[i * 3 + 1] = nDepth;
			pTrg[i * 3 + 2] = nDepth;
		}
	}
//		depthMap.setFromPixels(pDepth, camWidth, camHeight, OF_IMAGE_GRAYSCALE);
#endif
	depthMap.update();
	}
#else
    vidGrabber.update();

    if(vidGrabber.isFrameNew()){
        ofPixels & pixels = vidGrabber.getPixels();
        for(size_t i = 0; i < pixels.size(); i++){
            //invert the color of the pixel
            videoInverted[i] = 255 - pixels[i];
        }
        //load the inverted pixels
        videoTexture.loadData(videoInverted);
    }
#endif
}

//--------------------------------------------------------------
void ofApp::draw(){
    ofSetHexColor(0xffffff);
#if 1
    depthMap.draw(0, 0, ofGetWidth(), ofGetHeight());
	colorImg.draw(20, 20,camWidth/2,camHeight/2);
#else
    vidGrabber.draw(20, 20);
    videoTexture.draw(20 + camWidth, 20, camWidth, camHeight);
#endif
}


//--------------------------------------------------------------
void ofApp::keyPressed(int key){
    // in fullscreen mode, on a pc at least, the 
    // first time video settings the come up
    // they come up *under* the fullscreen window
    // use alt-tab to navigate to the settings
    // window. we are working on a fix for this...

    // Video settings no longer works in 10.7
    // You'll need to compile with the 10.6 SDK for this
    // For Xcode 4.4 and greater, see this forum post on instructions on installing the SDK
    // http://forum.openframeworks.cc/index.php?topic=10343
    if(key == 's' || key == 'S'){
        vidGrabber.videoSettings();
    }
    if(key == 'f' || key == 'F'){
        ofToggleFullscreen();
    }
}

//--------------------------------------------------------------
void ofApp::keyReleased(int key){
}

//--------------------------------------------------------------
void ofApp::mouseMoved(int x, int y){
}

//--------------------------------------------------------------
void ofApp::mouseDragged(int x, int y, int button){
}

//--------------------------------------------------------------
void ofApp::mousePressed(int x, int y, int button){
}

//--------------------------------------------------------------
void ofApp::mouseReleased(int x, int y, int button){
}

//--------------------------------------------------------------
void ofApp::mouseEntered(int x, int y){
}

//--------------------------------------------------------------
void ofApp::mouseExited(int x, int y){
}

//--------------------------------------------------------------
void ofApp::windowResized(int w, int h){
}

//--------------------------------------------------------------
void ofApp::gotMessage(ofMessage msg){
}

//--------------------------------------------------------------
void ofApp::dragEvent(ofDragInfo dragInfo){
}
